package java.util.stream;

import java.util.Spliterator;
import java.util.concurrent.atomic.AtomicReference;

abstract class AbstractShortCircuitTask<P_IN, P_OUT, R, K extends AbstractShortCircuitTask<P_IN, P_OUT, R, K>> extends AbstractTask<P_IN, P_OUT, R, K>
{
    protected final AtomicReference<R> sharedResult;
    protected volatile boolean canceled;
    
    protected AbstractShortCircuitTask(PipelineHelper<P_OUT> pipelineHelper, Spliterator<P_IN> spliterator) {
        super(pipelineHelper, spliterator);
        this.sharedResult = new AtomicReference(null);
    }
    
    protected AbstractShortCircuitTask(K k, Spliterator<P_IN> spliterator) {
        super(k, spliterator);
        this.sharedResult = k.sharedResult;
    }
    
    protected abstract R getEmptyResult();
    
    @Override
    public void compute() {
        Spliterator<P_IN> spliterator = this.spliterator;
        long n = spliterator.estimateSize();
        long targetSize = this.getTargetSize(n);
        int n2 = 0;
        Object localResult;
        while ((localResult = this.sharedResult.get()) == null) {
            if (this.taskCanceled()) {
                localResult = this.getEmptyResult();
                break;
            }
            Spliterator<P_IN> trySplit;
            if (n <= targetSize || (trySplit = spliterator.trySplit()) == null) {
                localResult = this.doLeaf();
                break;
            }
            AbstractShortCircuitTask<Object, Object, Object, Object> abstractShortCircuitTask = this.leftChild = (AbstractShortCircuitTask)this.makeChild(trySplit);
            AbstractShortCircuitTask<Object, Object, Object, Object> abstractShortCircuitTask2 = this.rightChild = (AbstractShortCircuitTask)this.makeChild(spliterator);
            this.setPendingCount(1);
            if (n2 != 0) {
                n2 = 0;
                spliterator = trySplit;
                this = abstractShortCircuitTask;
            }
            else {
                n2 = 1;
                this = abstractShortCircuitTask2;
                abstractShortCircuitTask2 = abstractShortCircuitTask;
            }
            abstractShortCircuitTask2.fork();
            n = spliterator.estimateSize();
        }
        this.setLocalResult(localResult);
        this.tryComplete();
    }
    
    protected void shortCircuit(R newValue) {
        if (newValue != null) {
            this.sharedResult.compareAndSet(null, newValue);
        }
    }
    
    @Override
    protected void setLocalResult(R r) {
        if (this.isRoot()) {
            if (r != null) {
                this.sharedResult.compareAndSet(null, r);
            }
        }
        else {
            super.setLocalResult(r);
        }
    }
    
    @Override
    public R getRawResult() {
        return this.getLocalResult();
    }
    
    public R getLocalResult() {
        if (this.isRoot()) {
            Object value = this.sharedResult.get();
            return (value == null) ? this.getEmptyResult() : value;
        }
        return super.getLocalResult();
    }
    
    protected void cancel() {
        this.canceled = true;
    }
    
    protected boolean taskCanceled() {
        int n = this.canceled;
        if (n == 0) {
            for (AbstractShortCircuitTask abstractShortCircuitTask = (AbstractShortCircuitTask)this.getParent(); n == 0 && abstractShortCircuitTask != null; n = abstractShortCircuitTask.canceled, abstractShortCircuitTask = (AbstractShortCircuitTask)abstractShortCircuitTask.getParent()) {}
        }
        return n;
    }
    
    protected void cancelLaterNodes() {
        AbstractShortCircuitTask<Object, Object, Object, Object> abstractShortCircuitTask = (AbstractShortCircuitTask)this.getParent();
        while (abstractShortCircuitTask != null) {
            if (abstractShortCircuitTask.leftChild == this) {
                AbstractShortCircuitTask<Object, Object, Object, Object> abstractShortCircuitTask2 = (AbstractShortCircuitTask)abstractShortCircuitTask.rightChild;
                if (!abstractShortCircuitTask2.canceled) {
                    abstractShortCircuitTask2.cancel();
                }
            }
            this = abstractShortCircuitTask;
            abstractShortCircuitTask = (AbstractShortCircuitTask)abstractShortCircuitTask.getParent();
        }
    }
}
